@@ -1,8 +1,7 @@
-use crate::tree::{ADD_TOKEN_SIZE, FINISH_NODE_SIZE, START_NODE_SIZE};
+use crate::tree::{EventIdx, EventKind, ADD_TOKEN_SIZE, FINISH_NODE_SIZE, START_NODE_SIZE};
 use crate::{SyntaxElement, SyntaxToken, SyntaxTree, TextRange, TreeConfig};
 use std::hash::Hash;
 use std::marker::PhantomData;
-use std::num::NonZeroU32;
 
 /// A handle to a specific node in a specific [`SyntaxTree`].
 ///
@@ -12,7 +11,7 @@ use std::num::NonZeroU32;
 /// other than the one this node is from.
 #[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
 pub struct SyntaxNode<C> {
-    idx: NonZeroU32,
+    idx: EventIdx,
     tree_id: u32,
     phantom: PhantomData<C>,
 }
@@ -21,30 +20,22 @@ static_assertions::assert_eq_size!(SyntaxNode<()>, Option<SyntaxNode<()>>, u64);
 
 impl<C: TreeConfig> SyntaxNode<C> {
     #[inline(always)]
-    pub(crate) unsafe fn new(idx: u32, tree_id: u32) -> Self {
-        Self {
-            idx: if cfg!(debug_assertions) {
-                NonZeroU32::new(idx).unwrap()
-            } else {
-                NonZeroU32::new_unchecked(idx)
-            },
-            tree_id,
-            phantom: PhantomData,
-        }
+    pub(crate) unsafe fn new(idx: EventIdx, tree_id: u32) -> Self {
+        Self { idx, tree_id, phantom: PhantomData }
     }
 
     /// Returns the kind of this node.
     pub fn kind(self, tree: &SyntaxTree<C>) -> C::NodeKind {
         self.verify_tree(tree);
-        unsafe { tree.get_start_node(self.idx.get()).0 }
+        unsafe { tree.get_start_node(self.idx).kind }
     }
 
     /// Returns an iterator over the direct child nodes and tokens of this node.
     pub fn children(self, tree: &SyntaxTree<C>) -> impl Iterator<Item = SyntaxElement<C>> + '_ {
         self.verify_tree(tree);
         Children {
-            idx: self.idx.get() + START_NODE_SIZE,
-            finish_idx: unsafe { tree.get_start_node(self.idx.get()).1 },
+            idx: self.idx + START_NODE_SIZE,
+            finish_idx: unsafe { tree.get_start_node(self.idx).finish_node_idx },
             tree,
             tree_id: self.tree_id,
         }
@@ -54,8 +45,8 @@ impl<C: TreeConfig> SyntaxNode<C> {
     pub fn child_nodes(self, tree: &SyntaxTree<C>) -> impl Iterator<Item = SyntaxNode<C>> + '_ {
         self.verify_tree(tree);
         ChildNodes {
-            idx: self.idx.get() + START_NODE_SIZE,
-            finish_idx: unsafe { tree.get_start_node(self.idx.get()).1 },
+            idx: self.idx + START_NODE_SIZE,
+            finish_idx: unsafe { tree.get_start_node(self.idx).finish_node_idx },
             tree,
             tree_id: self.tree_id,
         }
@@ -65,8 +56,8 @@ impl<C: TreeConfig> SyntaxNode<C> {
     pub fn child_tokens(self, tree: &SyntaxTree<C>) -> impl Iterator<Item = SyntaxToken<C>> + '_ {
         self.verify_tree(tree);
         ChildTokens {
-            idx: self.idx.get() + START_NODE_SIZE,
-            finish_idx: unsafe { tree.get_start_node(self.idx.get()).1 },
+            idx: self.idx + START_NODE_SIZE,
+            finish_idx: unsafe { tree.get_start_node(self.idx).finish_node_idx },
             tree,
             tree_id: self.tree_id,
         }
@@ -77,8 +68,8 @@ impl<C: TreeConfig> SyntaxNode<C> {
     pub fn descendants(self, tree: &SyntaxTree<C>) -> impl Iterator<Item = SyntaxElement<C>> + '_ {
         self.verify_tree(tree);
         Descendants {
-            idx: self.idx.get() + START_NODE_SIZE,
-            finish_idx: unsafe { tree.get_start_node(self.idx.get()).1 },
+            idx: self.idx + START_NODE_SIZE,
+            finish_idx: unsafe { tree.get_start_node(self.idx).finish_node_idx },
             tree,
             tree_id: self.tree_id,
         }
@@ -92,8 +83,8 @@ impl<C: TreeConfig> SyntaxNode<C> {
     ) -> impl Iterator<Item = SyntaxNode<C>> + '_ {
         self.verify_tree(tree);
         DescendantNodes {
-            idx: self.idx.get() + START_NODE_SIZE,
-            finish_idx: unsafe { tree.get_start_node(self.idx.get()).1 },
+            idx: self.idx + START_NODE_SIZE,
+            finish_idx: unsafe { tree.get_start_node(self.idx).finish_node_idx },
             tree,
             tree_id: self.tree_id,
         }
@@ -107,8 +98,8 @@ impl<C: TreeConfig> SyntaxNode<C> {
     ) -> impl Iterator<Item = SyntaxToken<C>> + '_ {
         self.verify_tree(tree);
         DescendantTokens {
-            idx: self.idx.get() + START_NODE_SIZE,
-            finish_idx: unsafe { tree.get_start_node(self.idx.get()).1 },
+            idx: self.idx + START_NODE_SIZE,
+            finish_idx: unsafe { tree.get_start_node(self.idx).finish_node_idx },
             tree,
             tree_id: self.tree_id,
         }
@@ -117,16 +108,16 @@ impl<C: TreeConfig> SyntaxNode<C> {
     /// Returns the range this node spans in the original input.
     pub fn range(self, tree: &SyntaxTree<C>) -> TextRange {
         self.verify_tree(tree);
-        let (_, _, start, end) = unsafe { tree.get_start_node(self.idx.get()) };
-        TextRange::new(start.into(), end.into())
+        let start_node = unsafe { tree.get_start_node(self.idx) };
+        TextRange::new(start_node.start.into(), start_node.end.into())
     }
 
     /// Returns the text of all the tokens this node contains.
     pub fn text(self, tree: &SyntaxTree<C>) -> &str {
         self.verify_tree(tree);
         unsafe {
-            let (_, _, start, end) = tree.get_start_node(self.idx.get());
-            tree.get_text(start, end)
+            let start_node = tree.get_start_node(self.idx);
+            tree.get_text(start_node.start, start_node.end)
         }
     }
 
@@ -140,8 +131,8 @@ impl<C: TreeConfig> SyntaxNode<C> {
 }
 
 struct Children<'a, C> {
-    idx: u32,
-    finish_idx: u32,
+    idx: EventIdx,
+    finish_idx: EventIdx,
     tree: &'a SyntaxTree<C>,
     tree_id: u32,
 }
@@ -150,32 +141,32 @@ impl<C: TreeConfig> Iterator for Children<'_, C> {
     type Item = SyntaxElement<C>;
 
     fn next(&mut self) -> Option<Self::Item> {
-        while self.idx < self.finish_idx {
-            unsafe {
-                if self.tree.is_start_node(self.idx) {
-                    let (_, finish_node_idx, _, _) = self.tree.get_start_node(self.idx);
-                    let element = SyntaxElement::Node(SyntaxNode::new(self.idx, self.tree_id));
-                    self.idx = finish_node_idx + FINISH_NODE_SIZE;
-                    return Some(element);
-                }
-
-                if self.tree.is_add_token(self.idx) {
-                    let element = SyntaxElement::Token(SyntaxToken::new(self.idx, self.tree_id));
-                    self.idx += ADD_TOKEN_SIZE;
-                    return Some(element);
-                }
-            }
-
-            unreachable!()
+        if self.finish_idx <= self.idx {
+            return None;
         }
 
-        None
+        unsafe {
+            match self.tree.event_kind(self.idx) {
+                EventKind::StartNode => {
+                    let finish_node_idx = self.tree.get_start_node(self.idx).finish_node_idx;
+                    let element = SyntaxElement::Node(SyntaxNode::new(self.idx, self.tree_id));
+                    self.idx = finish_node_idx + FINISH_NODE_SIZE;
+                    Some(element)
+                }
+                EventKind::AddToken => {
+                    let element = SyntaxElement::Token(SyntaxToken::new(self.idx, self.tree_id));
+                    self.idx += ADD_TOKEN_SIZE;
+                    Some(element)
+                }
+                EventKind::FinishNode => unreachable!(),
+            }
+        }
     }
 }
 
 struct ChildNodes<'a, C> {
-    idx: u32,
-    finish_idx: u32,
+    idx: EventIdx,
+    finish_idx: EventIdx,
     tree: &'a SyntaxTree<C>,
     tree_id: u32,
 }
@@ -186,20 +177,20 @@ impl<C: TreeConfig> Iterator for ChildNodes<'_, C> {
     fn next(&mut self) -> Option<Self::Item> {
         while self.idx < self.finish_idx {
             unsafe {
-                if self.tree.is_start_node(self.idx) {
-                    let (_, finish_node_idx, _, _) = self.tree.get_start_node(self.idx);
-                    let node = SyntaxNode::new(self.idx, self.tree_id);
-                    self.idx = finish_node_idx + FINISH_NODE_SIZE;
-                    return Some(node);
-                }
-
-                if self.tree.is_add_token(self.idx) {
-                    self.idx += ADD_TOKEN_SIZE;
-                    continue;
+                match self.tree.event_kind(self.idx) {
+                    EventKind::StartNode => {
+                        let finish_node_idx = self.tree.get_start_node(self.idx).finish_node_idx;
+                        let node = SyntaxNode::new(self.idx, self.tree_id);
+                        self.idx = finish_node_idx + FINISH_NODE_SIZE;
+                        return Some(node);
+                    }
+                    EventKind::AddToken => {
+                        self.idx += ADD_TOKEN_SIZE;
+                        continue;
+                    }
+                    EventKind::FinishNode => unreachable!(),
                 }
             }
-
-            unreachable!()
         }
 
         None
@@ -207,8 +198,8 @@ impl<C: TreeConfig> Iterator for ChildNodes<'_, C> {
 }
 
 struct ChildTokens<'a, C> {
-    finish_idx: u32,
-    idx: u32,
+    finish_idx: EventIdx,
+    idx: EventIdx,
     tree: &'a SyntaxTree<C>,
     tree_id: u32,
 }
@@ -219,20 +210,20 @@ impl<C: TreeConfig> Iterator for ChildTokens<'_, C> {
     fn next(&mut self) -> Option<Self::Item> {
         while self.idx < self.finish_idx {
             unsafe {
-                if self.tree.is_start_node(self.idx) {
-                    let (_, finish_node_idx, _, _) = self.tree.get_start_node(self.idx);
-                    self.idx = finish_node_idx + FINISH_NODE_SIZE;
-                    continue;
-                }
-
-                if self.tree.is_add_token(self.idx) {
-                    let token = SyntaxToken::new(self.idx, self.tree_id);
-                    self.idx += ADD_TOKEN_SIZE;
-                    return Some(token);
+                match self.tree.event_kind(self.idx) {
+                    EventKind::StartNode => {
+                        let finish_node_idx = self.tree.get_start_node(self.idx).finish_node_idx;
+                        self.idx = finish_node_idx + FINISH_NODE_SIZE;
+                        continue;
+                    }
+                    EventKind::AddToken => {
+                        let token = SyntaxToken::new(self.idx, self.tree_id);
+                        self.idx += ADD_TOKEN_SIZE;
+                        return Some(token);
+                    }
+                    EventKind::FinishNode => unreachable!(),
                 }
             }
-
-            unreachable!()
         }
 
         None
@@ -240,8 +231,8 @@ impl<C: TreeConfig> Iterator for ChildTokens<'_, C> {
 }
 
 struct Descendants<'a, C> {
-    finish_idx: u32,
-    idx: u32,
+    finish_idx: EventIdx,
+    idx: EventIdx,
     tree: &'a SyntaxTree<C>,
     tree_id: u32,
 }
@@ -252,25 +243,24 @@ impl<C: TreeConfig> Iterator for Descendants<'_, C> {
     fn next(&mut self) -> Option<Self::Item> {
         while self.idx < self.finish_idx {
             unsafe {
-                if self.tree.is_start_node(self.idx) {
-                    let element = SyntaxElement::Node(SyntaxNode::new(self.idx, self.tree_id));
-                    self.idx += START_NODE_SIZE;
-                    return Some(element);
-                }
-
-                if self.tree.is_add_token(self.idx) {
-                    let element = SyntaxElement::Token(SyntaxToken::new(self.idx, self.tree_id));
-                    self.idx += ADD_TOKEN_SIZE;
-                    return Some(element);
-                }
-
-                if self.tree.is_finish_node(self.idx) {
-                    self.idx += FINISH_NODE_SIZE;
-                    continue;
+                match self.tree.event_kind(self.idx) {
+                    EventKind::StartNode => {
+                        let element = SyntaxElement::Node(SyntaxNode::new(self.idx, self.tree_id));
+                        self.idx += START_NODE_SIZE;
+                        return Some(element);
+                    }
+                    EventKind::AddToken => {
+                        let element =
+                            SyntaxElement::Token(SyntaxToken::new(self.idx, self.tree_id));
+                        self.idx += ADD_TOKEN_SIZE;
+                        return Some(element);
+                    }
+                    EventKind::FinishNode => {
+                        self.idx += FINISH_NODE_SIZE;
+                        continue;
+                    }
                 }
             }
-
-            unreachable!()
         }
 
         None
@@ -278,8 +268,8 @@ impl<C: TreeConfig> Iterator for Descendants<'_, C> {
 }
 
 struct DescendantNodes<'a, C> {
-    finish_idx: u32,
-    idx: u32,
+    finish_idx: EventIdx,
+    idx: EventIdx,
     tree: &'a SyntaxTree<C>,
     tree_id: u32,
 }
@@ -290,24 +280,22 @@ impl<C: TreeConfig> Iterator for DescendantNodes<'_, C> {
     fn next(&mut self) -> Option<Self::Item> {
         while self.idx < self.finish_idx {
             unsafe {
-                if self.tree.is_start_node(self.idx) {
-                    let node = SyntaxNode::new(self.idx, self.tree_id);
-                    self.idx += START_NODE_SIZE;
-                    return Some(node);
-                }
-
-                if self.tree.is_add_token(self.idx) {
-                    self.idx += ADD_TOKEN_SIZE;
-                    continue;
-                }
-
-                if self.tree.is_finish_node(self.idx) {
-                    self.idx += FINISH_NODE_SIZE;
-                    continue;
+                match self.tree.event_kind(self.idx) {
+                    EventKind::StartNode => {
+                        let node = SyntaxNode::new(self.idx, self.tree_id);
+                        self.idx += START_NODE_SIZE;
+                        return Some(node);
+                    }
+                    EventKind::AddToken => {
+                        self.idx += ADD_TOKEN_SIZE;
+                        continue;
+                    }
+                    EventKind::FinishNode => {
+                        self.idx += FINISH_NODE_SIZE;
+                        continue;
+                    }
                 }
             }
-
-            unreachable!()
         }
 
         None
@@ -315,8 +303,8 @@ impl<C: TreeConfig> Iterator for DescendantNodes<'_, C> {
 }
 
 struct DescendantTokens<'a, C> {
-    finish_idx: u32,
-    idx: u32,
+    finish_idx: EventIdx,
+    idx: EventIdx,
     tree: &'a SyntaxTree<C>,
     tree_id: u32,
 }
@@ -327,24 +315,22 @@ impl<C: TreeConfig> Iterator for DescendantTokens<'_, C> {
     fn next(&mut self) -> Option<Self::Item> {
         while self.idx < self.finish_idx {
             unsafe {
-                if self.tree.is_add_token(self.idx) {
-                    let token = SyntaxToken::new(self.idx, self.tree_id);
-                    self.idx += ADD_TOKEN_SIZE;
-                    return Some(token);
-                }
-
-                if self.tree.is_start_node(self.idx) {
-                    self.idx += START_NODE_SIZE;
-                    continue;
-                }
-
-                if self.tree.is_finish_node(self.idx) {
-                    self.idx += FINISH_NODE_SIZE;
-                    continue;
+                match self.tree.event_kind(self.idx) {
+                    EventKind::StartNode => {
+                        self.idx += START_NODE_SIZE;
+                        continue;
+                    }
+                    EventKind::AddToken => {
+                        let token = SyntaxToken::new(self.idx, self.tree_id);
+                        self.idx += ADD_TOKEN_SIZE;
+                        return Some(token);
+                    }
+                    EventKind::FinishNode => {
+                        self.idx += FINISH_NODE_SIZE;
+                        continue;
+                    }
                 }
             }
-
-            unreachable!()
         }
 
         None
@@ -432,20 +418,20 @@ mod tests {
         let root = tree.root();
 
         let mut children = root.children(&tree);
-        let binary_expr = children.next().unwrap().assert_node();
+        let binary_expr = children.next().unwrap().unwrap_node();
         assert_eq!(binary_expr.kind(&tree), NodeKind::BinaryExpr);
-        let call = children.next().unwrap().assert_node();
+        let call = children.next().unwrap().unwrap_node();
         assert_eq!(call.kind(&tree), NodeKind::Call);
         assert!(children.next().is_none());
 
         let mut children = binary_expr.children(&tree);
-        assert_eq!(children.next().unwrap().assert_node().kind(&tree), NodeKind::BinaryExpr);
-        assert_eq!(children.next().unwrap().assert_token().kind(&tree), TokenKind::Plus);
-        assert_eq!(children.next().unwrap().assert_token().kind(&tree), TokenKind::IntLiteral);
+        assert_eq!(children.next().unwrap().unwrap_node().kind(&tree), NodeKind::BinaryExpr);
+        assert_eq!(children.next().unwrap().unwrap_token().kind(&tree), TokenKind::Plus);
+        assert_eq!(children.next().unwrap().unwrap_token().kind(&tree), TokenKind::IntLiteral);
         assert!(children.next().is_none());
 
         let mut children = call.children(&tree);
-        assert_eq!(children.next().unwrap().assert_token().kind(&tree), TokenKind::Ident);
+        assert_eq!(children.next().unwrap().unwrap_token().kind(&tree), TokenKind::Ident);
         assert!(children.next().is_none());
     }
 
@@ -500,21 +486,21 @@ mod tests {
         let root = tree.root();
 
         let mut descendants = root.descendants(&tree);
-        let binary_expr = descendants.next().unwrap().assert_node();
+        let binary_expr = descendants.next().unwrap().unwrap_node();
         assert_eq!(binary_expr.kind(&tree), NodeKind::BinaryExpr);
 
-        let binary_expr_2 = descendants.next().unwrap().assert_node();
+        let binary_expr_2 = descendants.next().unwrap().unwrap_node();
         assert_eq!(binary_expr_2.kind(&tree), NodeKind::BinaryExpr);
-        assert_eq!(descendants.next().unwrap().assert_token().kind(&tree), TokenKind::IntLiteral);
-        assert_eq!(descendants.next().unwrap().assert_token().kind(&tree), TokenKind::Asterisk);
-        assert_eq!(descendants.next().unwrap().assert_token().kind(&tree), TokenKind::IntLiteral);
+        assert_eq!(descendants.next().unwrap().unwrap_token().kind(&tree), TokenKind::IntLiteral);
+        assert_eq!(descendants.next().unwrap().unwrap_token().kind(&tree), TokenKind::Asterisk);
+        assert_eq!(descendants.next().unwrap().unwrap_token().kind(&tree), TokenKind::IntLiteral);
 
-        assert_eq!(descendants.next().unwrap().assert_token().kind(&tree), TokenKind::Plus);
-        assert_eq!(descendants.next().unwrap().assert_token().kind(&tree), TokenKind::IntLiteral);
+        assert_eq!(descendants.next().unwrap().unwrap_token().kind(&tree), TokenKind::Plus);
+        assert_eq!(descendants.next().unwrap().unwrap_token().kind(&tree), TokenKind::IntLiteral);
 
-        let call = descendants.next().unwrap().assert_node();
+        let call = descendants.next().unwrap().unwrap_node();
         assert_eq!(call.kind(&tree), NodeKind::Call);
-        assert_eq!(descendants.next().unwrap().assert_token().kind(&tree), TokenKind::Ident);
+        assert_eq!(descendants.next().unwrap().unwrap_token().kind(&tree), TokenKind::Ident);
         assert!(descendants.next().is_none());
 
         let mut descendants = binary_expr.child_nodes(&tree);
