@@ -4,6 +4,8 @@ use self::tag::Tag;
 use crate::{SyntaxNode, SyntaxToken, TextRange, TreeConfig};
 use std::fmt;
 use std::marker::PhantomData;
+use std::num::NonZeroU32;
+use std::ops::{Add, AddAssign};
 use std::sync::atomic::{AtomicU32, Ordering};
 
 /// `SyntaxTree` owns the syntax tree allocation.
@@ -27,7 +29,7 @@ use std::sync::atomic::{AtomicU32, Ordering};
 /// - `[u8]` events
 ///
 /// These are stored contiguously in one memory allocation.
-/// Nodes and tokens are a `u32` index into this allocation.
+/// Nodes and tokens are a `u32` byte index into this allocation.
 /// All numerical types are stored in the target platform’s native endianness.
 ///
 /// ## ID
@@ -101,13 +103,16 @@ pub struct SyntaxBuilder<C> {
     phantom: PhantomData<C>,
 }
 
-pub(crate) const START_NODE_SIZE: u32 = 2 + 4 + 4 + 4;
-pub(crate) const ADD_TOKEN_SIZE: u32 = 2 + 4 + 4;
-pub(crate) const FINISH_NODE_SIZE: u32 = 2;
+pub(crate) const START_NODE_SIZE: EventSize = EventSize(2 + 4 + 4 + 4);
+pub(crate) const ADD_TOKEN_SIZE: EventSize = EventSize(2 + 4 + 4);
+pub(crate) const FINISH_NODE_SIZE: EventSize = EventSize(2);
 
 const FINISH_NODE_IDX_PLACEHOLDER: u32 = 0;
 
-static CURRENT_TREE_ID: AtomicU32 = AtomicU32::new(0);
+fn gen_tree_id() -> u32 {
+    static CURRENT: AtomicU32 = AtomicU32::new(0);
+    CURRENT.fetch_add(1, Ordering::Relaxed)
+}
 
 impl<C: TreeConfig> SyntaxBuilder<C> {
     /// Constructs a new empty `SyntaxBuilder` with the provided source text.
@@ -129,12 +134,14 @@ impl<C: TreeConfig> SyntaxBuilder<C> {
     ) -> Self {
         assert!(text.len() < u32::MAX as usize);
 
-        let id = CURRENT_TREE_ID.fetch_add(1, Ordering::SeqCst);
+        let id = gen_tree_id();
 
         let mut data = Vec::with_capacity(
-            start_nodes * START_NODE_SIZE as usize
-                + add_tokens * ADD_TOKEN_SIZE as usize
-                + finish_nodes * FINISH_NODE_SIZE as usize,
+            4 + 4
+                + text.len()
+                + start_nodes * START_NODE_SIZE.to_usize()
+                + add_tokens * ADD_TOKEN_SIZE.to_usize()
+                + finish_nodes * FINISH_NODE_SIZE.to_usize(),
         );
 
         data.extend_from_slice(&id.to_ne_bytes());
@@ -168,14 +175,16 @@ impl<C: TreeConfig> SyntaxBuilder<C> {
 
         self.start_node_idxs.push(self.data.len());
 
-        self.data.reserve(START_NODE_SIZE as usize);
+        self.data.reserve(START_NODE_SIZE.to_usize());
         unsafe {
-            let ptr = self.data_end_ptr();
-            (ptr as *mut Tag).write_unaligned(Tag::start_node::<C>(kind));
-            (ptr.add(2) as *mut u32).write_unaligned(FINISH_NODE_IDX_PLACEHOLDER);
-            (ptr.add(6) as *mut u32).write_unaligned(self.current_len);
-            (ptr.add(10) as *mut u32).write_unaligned(self.current_len);
-            self.data.set_len(self.data.len() + START_NODE_SIZE as usize);
+            (self.data.as_mut_ptr_range().end as *mut RawStartNode).write_unaligned(RawStartNode {
+                tag: Tag::start_node::<C>(kind),
+                finish_node_idx: FINISH_NODE_IDX_PLACEHOLDER,
+                start: self.current_len,
+                end: self.current_len,
+            });
+
+            self.data.set_len(self.data.len() + START_NODE_SIZE.to_usize());
         }
     }
 
@@ -207,13 +216,16 @@ impl<C: TreeConfig> SyntaxBuilder<C> {
         let end = u32::from(range.end());
         self.current_len = end;
 
-        self.data.reserve(ADD_TOKEN_SIZE as usize);
+        self.data.reserve(ADD_TOKEN_SIZE.to_usize());
+
         unsafe {
-            let ptr = self.data_end_ptr();
-            (ptr as *mut Tag).write_unaligned(Tag::add_token::<C>(kind));
-            (ptr.add(2) as *mut u32).write_unaligned(start);
-            (ptr.add(6) as *mut u32).write_unaligned(end);
-            self.data.set_len(self.data.len() + ADD_TOKEN_SIZE as usize);
+            (self.data.as_mut_ptr_range().end as *mut RawAddToken).write_unaligned(RawAddToken {
+                tag: Tag::add_token::<C>(kind),
+                start,
+                end,
+            });
+
+            self.data.set_len(self.data.len() + ADD_TOKEN_SIZE.to_usize());
         }
     }
 
@@ -230,24 +242,24 @@ impl<C: TreeConfig> SyntaxBuilder<C> {
         let start_node_idx = self.start_node_idxs.pop().unwrap();
         let finish_node_idx = self.data.len() as u32;
 
-        self.data.reserve(FINISH_NODE_SIZE as usize);
+        self.data.reserve(FINISH_NODE_SIZE.to_usize());
         unsafe {
-            let ptr = self.data_end_ptr() as *mut Tag;
+            let ptr = self.data.as_mut_ptr_range().end as *mut Tag;
             ptr.write_unaligned(Tag::finish_node());
-            self.data.set_len(self.data.len() + FINISH_NODE_SIZE as usize);
+            self.data.set_len(self.data.len() + FINISH_NODE_SIZE.to_usize());
         }
 
         unsafe {
-            let ptr = self.data.as_mut_ptr().add(start_node_idx);
-            debug_assert!((ptr as *const Tag).read_unaligned().is_start_node());
+            let ptr = &mut *(self.data.as_mut_ptr().add(start_node_idx) as *mut RawStartNode);
+            debug_assert_eq!(ptr.tag.event_kind(), EventKind::StartNode);
 
-            debug_assert_eq!(
-                (ptr.add(2) as *const u32).read_unaligned(),
-                FINISH_NODE_IDX_PLACEHOLDER
-            );
-            (ptr.add(2) as *mut u32).write_unaligned(finish_node_idx);
+            // debug_assert_eq tries to take a reference to the field,
+            // which isn’t allowed since it’s packed,
+            // so we use a manual debug_assert instead
+            debug_assert!(ptr.finish_node_idx == FINISH_NODE_IDX_PLACEHOLDER);
 
-            (ptr.add(10) as *mut u32).write_unaligned(self.current_len);
+            ptr.finish_node_idx = finish_node_idx;
+            ptr.end = self.current_len;
         }
     }
 
@@ -273,21 +285,20 @@ impl<C: TreeConfig> SyntaxBuilder<C> {
         let len = self.text_len() as usize;
         unsafe {
             let s = self.data.get_unchecked(8..len + 8);
-            if cfg!(debug_assertions) {
-                std::str::from_utf8(s).unwrap()
-            } else {
-                std::str::from_utf8_unchecked(s)
-            }
+
+            // has to stay unchecked even in debug mode
+            // since this method is called every time a token is added
+            //
+            // if we perform an operation in this method that depends on the input size,
+            // then tree construction becomes O(n^2)
+            // (since input size and the number of tokens are roughly proportional)
+            std::str::from_utf8_unchecked(s)
         }
     }
 
     fn text_len(&self) -> u32 {
         unsafe { (self.data.as_ptr() as *const u32).add(1).read_unaligned() }
     }
-
-    fn data_end_ptr(&mut self) -> *mut u8 {
-        unsafe { self.data.as_mut_ptr().add(self.data.len()) }
-    }
 }
 
 impl<C: TreeConfig> SyntaxTree<C> {
@@ -321,9 +332,11 @@ impl<C: TreeConfig> SyntaxTree<C> {
         RawEvents { idx: self.root_idx(), tree: self }
     }
 
-    pub(crate) fn root_idx(&self) -> u32 {
-        let text_len = unsafe { (self.data.as_ptr() as *const u32).add(1).read_unaligned() };
-        text_len + 8
+    pub(crate) fn root_idx(&self) -> EventIdx {
+        unsafe {
+            let text_len = (self.data.as_ptr() as *const u32).add(1).read_unaligned();
+            EventIdx::new(text_len + 8)
+        }
     }
 
     pub(crate) fn id(&self) -> u32 {
@@ -343,56 +356,117 @@ impl<C: TreeConfig> SyntaxTree<C> {
         }
     }
 
-    pub(crate) unsafe fn get_start_node(&self, idx: u32) -> (C::NodeKind, u32, u32, u32) {
-        let idx = idx as usize;
-        debug_assert!(idx + START_NODE_SIZE as usize <= self.data.len());
+    pub(crate) unsafe fn get_start_node(&self, idx: EventIdx) -> StartNode<C> {
+        let idx = idx.0.get() as usize;
+        debug_assert!(idx + START_NODE_SIZE.to_usize() <= self.data.len());
 
-        let ptr = self.data.as_ptr().add(idx);
-        let tag = (ptr as *const Tag).read_unaligned();
-        let finish_node_idx = (ptr.add(2) as *const u32).read_unaligned();
-        let start = (ptr.add(6) as *const u32).read_unaligned();
-        let end = (ptr.add(10) as *const u32).read_unaligned();
+        let ptr = self.data.as_ptr().add(idx) as *const RawStartNode;
+        let raw = ptr.read_unaligned();
 
-        let kind = tag.get_start_node_kind::<C>();
-
-        (kind, finish_node_idx, start, end)
+        StartNode {
+            kind: raw.tag.get_start_node_kind::<C>(),
+            finish_node_idx: EventIdx::new(raw.finish_node_idx),
+            start: raw.start,
+            end: raw.end,
+        }
     }
 
-    pub(crate) unsafe fn get_add_token(&self, idx: u32) -> (C::TokenKind, u32, u32) {
-        let idx = idx as usize;
-        debug_assert!(idx + ADD_TOKEN_SIZE as usize <= self.data.len());
+    pub(crate) unsafe fn get_add_token(&self, idx: EventIdx) -> AddToken<C> {
+        let idx = idx.0.get() as usize;
+        debug_assert!(idx + ADD_TOKEN_SIZE.to_usize() <= self.data.len());
 
-        let ptr = self.data.as_ptr().add(idx);
-        let tag = (ptr as *const Tag).read_unaligned();
-        let start = (ptr.add(2) as *const u32).read_unaligned();
-        let end = (ptr.add(6) as *const u32).read_unaligned();
+        let ptr = self.data.as_ptr().add(idx) as *const RawAddToken;
+        let raw = ptr.read_unaligned();
 
-        let kind = tag.get_add_token_kind::<C>();
-
-        (kind, start, end)
+        AddToken { kind: raw.tag.get_add_token_kind::<C>(), start: raw.start, end: raw.end }
     }
 
-    pub(crate) unsafe fn is_start_node(&self, idx: u32) -> bool {
-        self.tag_at_idx(idx).is_start_node()
+    pub(crate) unsafe fn event_kind(&self, idx: EventIdx) -> EventKind {
+        self.tag_at_idx(idx).event_kind()
     }
 
-    pub(crate) unsafe fn is_add_token(&self, idx: u32) -> bool {
-        self.tag_at_idx(idx).is_add_token()
-    }
-
-    pub(crate) unsafe fn is_finish_node(&self, idx: u32) -> bool {
-        self.tag_at_idx(idx).is_finish_node()
-    }
-
-    fn tag_at_idx(&self, idx: u32) -> Tag {
-        let idx = idx as usize;
+    fn tag_at_idx(&self, idx: EventIdx) -> Tag {
+        let idx = idx.0.get() as usize;
         debug_assert!(idx < self.data.len());
         unsafe { (self.data.as_ptr().add(idx) as *const Tag).read_unaligned() }
     }
 }
 
+#[repr(C, packed)]
+pub(crate) struct StartNode<C: TreeConfig> {
+    pub(crate) kind: C::NodeKind,
+    pub(crate) finish_node_idx: EventIdx,
+    pub(crate) start: u32,
+    pub(crate) end: u32,
+}
+
+#[repr(C, packed)]
+struct RawStartNode {
+    tag: Tag,
+    finish_node_idx: u32,
+    start: u32,
+    end: u32,
+}
+
+#[repr(C, packed)]
+pub(crate) struct AddToken<C: TreeConfig> {
+    pub(crate) kind: C::TokenKind,
+    pub(crate) start: u32,
+    pub(crate) end: u32,
+}
+
+#[repr(C, packed)]
+struct RawAddToken {
+    tag: Tag,
+    start: u32,
+    end: u32,
+}
+
+#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
+pub(crate) struct EventIdx(NonZeroU32);
+
+impl EventIdx {
+    pub(crate) unsafe fn new(idx: u32) -> Self {
+        if cfg!(debug_assertions) {
+            Self(NonZeroU32::new(idx).unwrap())
+        } else {
+            Self(NonZeroU32::new_unchecked(idx))
+        }
+    }
+}
+
+#[derive(Clone, Copy)]
+pub(crate) struct EventSize(u32);
+
+impl EventSize {
+    fn to_usize(self) -> usize {
+        self.0 as usize
+    }
+}
+
+impl Add<EventSize> for EventIdx {
+    type Output = Self;
+
+    fn add(self, rhs: EventSize) -> Self::Output {
+        unsafe { Self::new(self.0.get() + rhs.0) }
+    }
+}
+
+impl AddAssign<EventSize> for EventIdx {
+    fn add_assign(&mut self, rhs: EventSize) {
+        *self = *self + rhs;
+    }
+}
+
+#[derive(Debug, PartialEq)]
+pub(crate) enum EventKind {
+    StartNode,
+    AddToken,
+    FinishNode,
+}
+
 struct Events<'a, C> {
-    idx: u32,
+    idx: EventIdx,
     tree: &'a SyntaxTree<C>,
 }
 
@@ -400,33 +474,31 @@ impl<C: TreeConfig> Iterator for Events<'_, C> {
     type Item = Event<C>;
 
     fn next(&mut self) -> Option<Self::Item> {
-        if self.idx >= self.tree.data.len() as u32 {
+        if self.idx.0.get() >= self.tree.data.len() as u32 {
             return None;
         }
 
-        if unsafe { self.tree.is_start_node(self.idx) } {
-            let node = unsafe { SyntaxNode::new(self.idx, self.tree.id()) };
-            self.idx += START_NODE_SIZE;
-            return Some(Event::StartNode(node));
+        match unsafe { self.tree.event_kind(self.idx) } {
+            EventKind::StartNode => {
+                let node = unsafe { SyntaxNode::new(self.idx, self.tree.id()) };
+                self.idx += START_NODE_SIZE;
+                Some(Event::StartNode(node))
+            }
+            EventKind::AddToken => {
+                let token = unsafe { SyntaxToken::new(self.idx, self.tree.id()) };
+                self.idx += ADD_TOKEN_SIZE;
+                Some(Event::AddToken(token))
+            }
+            EventKind::FinishNode => {
+                self.idx += FINISH_NODE_SIZE;
+                Some(Event::FinishNode)
+            }
         }
-
-        if unsafe { self.tree.is_add_token(self.idx) } {
-            let token = unsafe { SyntaxToken::new(self.idx, self.tree.id()) };
-            self.idx += ADD_TOKEN_SIZE;
-            return Some(Event::AddToken(token));
-        }
-
-        if unsafe { self.tree.is_finish_node(self.idx) } {
-            self.idx += FINISH_NODE_SIZE;
-            return Some(Event::FinishNode);
-        }
-
-        unreachable!()
     }
 }
 
 struct RawEvents<'a, C> {
-    idx: u32,
+    idx: EventIdx,
     tree: &'a SyntaxTree<C>,
 }
 
@@ -434,30 +506,28 @@ impl<C: TreeConfig> Iterator for RawEvents<'_, C> {
     type Item = RawEvent<C>;
 
     fn next(&mut self) -> Option<Self::Item> {
-        if self.idx >= self.tree.data.len() as u32 {
+        if self.idx.0.get() >= self.tree.data.len() as u32 {
             return None;
         }
 
-        if unsafe { self.tree.is_start_node(self.idx) } {
-            let (kind, _, start, end) = unsafe { self.tree.get_start_node(self.idx) };
-            let range = TextRange::new(start.into(), end.into());
-            self.idx += START_NODE_SIZE;
-            return Some(RawEvent::StartNode { kind, range });
+        match unsafe { self.tree.event_kind(self.idx) } {
+            EventKind::StartNode => {
+                let start_node = unsafe { self.tree.get_start_node(self.idx) };
+                let range = TextRange::new(start_node.start.into(), start_node.end.into());
+                self.idx += START_NODE_SIZE;
+                Some(RawEvent::StartNode { kind: start_node.kind, range })
+            }
+            EventKind::AddToken => {
+                let add_token = unsafe { self.tree.get_add_token(self.idx) };
+                let range = TextRange::new(add_token.start.into(), add_token.end.into());
+                self.idx += ADD_TOKEN_SIZE;
+                Some(RawEvent::AddToken { kind: add_token.kind, range })
+            }
+            EventKind::FinishNode => {
+                self.idx += FINISH_NODE_SIZE;
+                Some(RawEvent::FinishNode)
+            }
         }
-
-        if unsafe { self.tree.is_add_token(self.idx) } {
-            let (kind, start, end) = unsafe { self.tree.get_add_token(self.idx) };
-            let range = TextRange::new(start.into(), end.into());
-            self.idx += ADD_TOKEN_SIZE;
-            return Some(RawEvent::AddToken { kind, range });
-        }
-
-        if unsafe { self.tree.is_finish_node(self.idx) } {
-            self.idx += FINISH_NODE_SIZE;
-            return Some(RawEvent::FinishNode);
-        }
-
-        unreachable!()
     }
 }
 
