@@ -1,3 +1,4 @@
+use super::EventKind;
 use crate::{SyntaxKind, TreeConfig};
 
 #[derive(Clone, Copy)]
@@ -23,27 +24,27 @@ impl Tag {
         Self(u16::MAX)
     }
 
-    pub(super) fn is_start_node(self) -> bool {
-        self.high_bit_is_1() && !self.is_finish_node()
-    }
-
-    pub(super) fn is_add_token(self) -> bool {
-        !self.high_bit_is_1()
-    }
-
-    pub(super) fn is_finish_node(self) -> bool {
-        self.0 == u16::MAX
+    pub(super) fn event_kind(self) -> EventKind {
+        if self.high_bit_is_1() {
+            if self.0 == u16::MAX {
+                EventKind::FinishNode
+            } else {
+                EventKind::StartNode
+            }
+        } else {
+            EventKind::AddToken
+        }
     }
 
     pub(super) fn get_start_node_kind<C: TreeConfig>(self) -> C::NodeKind {
-        debug_assert!(self.is_start_node());
+        debug_assert_eq!(self.event_kind(), EventKind::StartNode);
         let raw = self.0 & u16::MAX >> 1; // zero out high bit
         debug_assert!(raw <= Self::MAX_KIND);
         unsafe { C::NodeKind::from_raw(raw) }
     }
 
     pub(super) fn get_add_token_kind<C: TreeConfig>(self) -> C::TokenKind {
-        debug_assert!(self.is_add_token());
+        debug_assert_eq!(self.event_kind(), EventKind::AddToken);
         debug_assert!(self.0 <= Self::MAX_KIND);
         unsafe { C::TokenKind::from_raw(self.0) }
     }
